#include "config.h"

#include <aviutil.h>
#include <time.h>
#include <registry.h>
#include <wine/winreg.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <default.h>

#include <sys/time.h>
#include <unistd.h>
unsigned int (*localcount)();
long long (*longcount)();

static void do_cpuid(unsigned int *regs)
{
	unsigned int ax;
	ax=1;
	__asm__ __volatile__(
	"pushl %%ebx; pushl %%ecx; pushl %%edx;"
	".byte  0x0f, 0xa2;"
	"movl   %%eax, (%2);"
	"movl   %%ebx, 4(%2);"
	"movl   %%ecx, 8(%2);"
	"movl   %%edx, 12(%2);"
	"popl %%edx; popl %%ecx; popl %%ebx;"
	: "=a" (ax)
	:  "0" (ax), "S" (regs)
	);
	return;
}

unsigned int CPU_Info::localcount_tsc()
{
    int a;
    __asm__ __volatile__("rdtsc\n\t"
    :"=a"(a)
    :
    :"edx");
    return a;
}
long long CPU_Info::longcount_tsc()
{
    unsigned int i;
    unsigned int j;
    __asm__ __volatile__("rdtsc\n\t":"=a"(i), "=d"(j):);
    return ((long long)j<<32)+(long long)i;
}

unsigned int CPU_Info::localcount_notsc()
{
    struct timeval tv;
    unsigned limit=~0;
    limit/=1000000;
    gettimeofday(&tv, 0);
    return limit*tv.tv_usec;
}
long long CPU_Info::longcount_notsc()
{
    struct timeval tv;
    unsigned limit=~0;
    limit/=1000000;
    gettimeofday(&tv, 0);
    unsigned long long result;
    result=tv.tv_sec;
    result<<=32;
    result+=limit*tv.tv_usec;
    return result;
}


double CPU_Info::old_freq()
{
    int i=time(NULL);
    long long x,y;
    while(i==time(NULL));
    x=longcount();
    i++;
    while(i==time(NULL));
    y=longcount();
    return (double)(y-x)/1000.;
}

CPU_Info::CPU_Info() :have_mmx(0), have_tsc(0), freq(-1)
{
	char line[200];
	char model[200]="unknown";
	char flags[500]="";
	char	*s,*value;

	FILE *f = fopen ("/proc/cpuinfo", "r");
	if (!f)
	{
    	    unsigned int regs[4];
	    do_cpuid(regs);
	    have_tsc=regs[3] & 0x00000010;
	    have_mmx=regs[3] & 0x00800000;

    	    if(have_tsc)
	    {
		longcount=longcount_tsc;
		localcount=localcount_tsc;
	    }
	    else
	    {
	        longcount=longcount_notsc;
		localcount=localcount_notsc;
	    }
	    
	    freq=old_freq();
	    return;
	}

	while (fgets(line,200,f)!=NULL) 
	{

		/* NOTE: the ':' is the only character we can rely on */
		if (!(value = strchr(line,':')))
			continue;
		/* terminate the valuename */
		*value++ = '\0';
		/* skip any leading spaces */
		while (*value==' ') value++;
		if ((s=strchr(value,'\n')))
			*s='\0';

		if (!strncasecmp(line, "cpu MHz",strlen("cpu MHz"))) 
		{
		    freq=atof(value);
		    freq*=1000;
		}
		if(!strncasecmp(line, "model name", strlen("model name")))
		    strncpy(model, value, sizeof(model)-1);
		if((!strncasecmp(line, "flags", strlen("flags"))) ||
		(!strncasecmp(line, "features", strlen("features"))))
		    strncpy(flags, value, sizeof(flags)-1);
		continue;
		
	}
        cout<<"Available CPU flags: "<<flags<<endl;	
	if(strstr(flags, "tsc")==0)
	    have_tsc=0;
	else
	    have_tsc=1;
	if(strstr(flags, "mmx")==0)
	    have_mmx=0;
	else
	    have_mmx=1;

	if(have_tsc)
	{
	    longcount=longcount_tsc;
	    localcount=localcount_tsc;
	}
	else
	{
	    longcount=longcount_notsc;
	    localcount=localcount_notsc;
	}
	
	if(freq<0)freq=old_freq();
	if(have_tsc)
	    cout<<freq/1000.<<" MHz "<<model<<" processor detected"<<endl; 
	fclose(f);
}

CPU_Info freq;

int Registry::WriteInt(string appname, string valname, int value)
{
    return WriteData(appname, valname, &value, 4);
}

int Registry::ReadInt(string appname, string valname, int def_value)
{
    int res=def_value;
    if(ReadData(appname, valname, &res, 4)!=0)
	WriteInt(appname, valname, def_value);
    return res;
}
int Registry::WriteFloat(string appname, string valname, float value)
{
    return WriteData(appname, valname, &value, 4);
}

float Registry::ReadFloat(string appname, string valname, float def_value)
{
    float res=def_value;
    if(ReadData(appname, valname, &res, 4)!=0)
	WriteFloat(appname, valname, def_value);
    return res;
}

int Registry::WriteString(string appname, string valname, string value)
{
    return WriteData(appname, valname, value.c_str(), value.length()+1);
}

string Registry::ReadString(string appname, string valname, string def_value)
{
    char name[256];
    int result=ReadData(appname, valname, name, 256);
    if(result!=0)
    {
	WriteString(appname, valname, def_value);
	return def_value;
    }	
    name[255]=0;
    return string(name);
}

int Registry::WriteData(string appname, string valname, const void* data, int size)
{
    int result, status, newkey;
//    cerr<<appname<<" "<<valname<<endl;
    string fullname=string("Software\\Registry\\")+appname;
//    cerr<<fullname<<" qqq"<<endl;
    result=RegCreateKeyExA(HKEY_CURRENT_USER, fullname.c_str(), 0, 0, 0, 0, 0,
	&newkey, &status);
    if(result!=0)
        return -1;
    result=RegSetValueExA(newkey, valname.c_str(), 0, REG_DWORD, (const int*)data, size);
    RegCloseKey(newkey);
    return 0;
}

int Registry::ReadData(string appname, string valname, void* data, int size)
{
    int result, status, newkey;
    if(!data)return -1;
//    cerr<<appname<<" "<<valname<<endl;
    string fullname=string("Software\\Registry\\");
//    cerr<<fullname<<" zzz"<<endl;
    fullname+=appname;
//    cerr<<fullname<<" zzz"<<endl;
    result=RegCreateKeyExA(HKEY_CURRENT_USER, fullname.c_str(), 0, 0, 0, 0, 0,
    	&newkey, &status);
    if(result!=0)
        return -1;
    if(status==REG_CREATED_NEW_KEY)
    {
        RegCloseKey(newkey);
	return -1;
    }
    int count=size;
    result=RegQueryValueExA(newkey, valname.c_str(), 0, 0, (int*)data, &count)
;
    RegCloseKey(newkey);
    return result;
}
