#include "Module.h"
#include <loader.h>
#include <wine/winreg.h>
#include <registry.h>
#include <except.h>

#ifdef __FreeBSD__
#include <sys/time.h>
#endif
#define __MODULE__ "Win32 loader"

/*
#define STORE_ALL \
    __asm__ __volatile__ ( \
    "push %ebx\n\t" \
    "push %ecx\n\t" \
    "push %edx\n\t" \
    "push %esi\n\t" \
    "push %edi\n\t": :)
    
#define REST_ALL \
    __asm__ __volatile__ ( \
    "pop %edi\n\t" \
    "pop %esi\n\t" \
    "pop %edx\n\t" \
    "pop %ecx\n\t" \
    "pop %ebx\n\t": :)
*/
#define STORE_ALL
#define REST_ALL
using namespace std;
extern "C" char* def_path;

VideoCodecControl control;

Module::Module(string name, VideoCodecControl& parent)
    :_parent(parent), _name(name), _refcount(0), forgotten(0)
{
    _fullname=def_path;
    _fullname+="/";
    _fullname+=name;
    _handle= LoadLibraryA(_fullname.c_str());
    if(_handle==0)
	throw FATAL("Could not load library");
    try
    {	
	if((name==string("divxc32.dll")) || (name==string("divxc32f.dll")))
	{
	    int newkey;
	    int bitrate;	
    	    int count=4;
	    if(RegOpenKeyExA(HKEY_CURRENT_USER, "Software\\LinuxLoader\\div3", 0, 0, &newkey)!=0)
	    {
//		cerr<<"Could not open Divx key"<<endl;
		goto no_reg;
	    }	
	    if(RegQueryValueExA(newkey, "BitRate", 0, 0, &bitrate, &count)!=0)
	    {
//		cerr<<"No such value BitRate"<<endl;
		RegCloseKey(newkey);
	        goto no_reg;
    	    }    
 
	    *(double*)((char*)_handle+0x14c0)=bitrate;
	    RegCloseKey(newkey);
	no_reg:
	    ;	
	}	
   
	DriverProc=(DRIVER_PROC) GetProcAddress(_handle, "DriverProc");    
	if(!DriverProc)
	    throw FATAL("Not a valid Win32 codec");

	driver.uDriverSignature=0;
	driver.hDriverModule=_handle;
	driver.DriverProc=(DRIVERPROC)DriverProc;
	driver.dwDriverID=0;


	STORE_ALL;
        DriverProc(0, (int)&driver, DRV_LOAD, 0, 0);
	REST_ALL;

	STORE_ALL;
	DriverProc(0, (int)&driver, DRV_ENABLE, 0, 0);
	REST_ALL;
    }
    catch(FatalError& e)
    {
	FreeLibrary(_handle);
	throw;
    }	
}
int Module::CreateHandle(int compressor, Mode mode)
{
    ICOPEN icopen;
    icopen.fccType	= mmioFOURCC('v', 'i', 'd', 'c');
    icopen.fccHandler	= compressor;
    icopen.dwSize	= sizeof(ICOPEN);
    icopen.dwFlags	= (mode==Compress)?ICMODE_COMPRESS:ICMODE_DECOMPRESS;

    driver.dwDriverID=++_refcount;
    DRVR* hDriver=new DRVR(driver);
//    cerr<<"Creating new handle for _refcount "<<_refcount<<", driver "<<(int)hDriver<<endl;
//    cerr.setf(ios::hex, ios::basefield);
//    cerr<<"fcc: "<<icopen.fccType<<", handler: "<<icopen.fccHandler<<endl;
//    cerr.setf(ios::dec, ios::basefield);
    STORE_ALL;
    hDriver->dwDriverID=DriverProc((int)_refcount, (int)hDriver, DRV_OPEN,
         0, (int) &icopen);
    REST_ALL;
//    cerr<<"New ID "<<hDriver->dwDriverID<<endl;
    if(!hDriver->dwDriverID)
	throw FATAL("DRV_OPEN failed");
    return (int)hDriver;
}

int Module::CloseHandle(int handle)
{
    DRVR* hDriver=(DRVR*)handle;
    DriverProc(hDriver->dwDriverID, (int)hDriver, DRV_CLOSE, 0, 0);
    _refcount--;
    if(_refcount==0)
	delete this;
    delete hDriver;
    return 0;
}    
Module::~Module()
{
//    cerr<<"Destroying Module"<<endl;
    DriverProc(0, (int)&driver, DRV_FREE, 0, 0);
    FreeLibrary(_handle);
    if(!forgotten)_parent.Erase(this);
//    cerr<<"Destroying Module successful"<<endl;
}
int Module::Message(int handle, int message, int arg1, int arg2)
{
    int result;
    DRVR* module=(DRVR*)handle;
    char qw[200];    
#ifdef __FreeBSD__
	/*   XXX
	 *	disable Thread Context Switch in Windows Dll Calls.
	 *	Because Thread Switched Win32 Dll Calls <-> another threads
	 *      broken cpu registers ???
	 */
    struct itimerval itoval;
    struct itimerval itval;
    memset(&itval,0,sizeof(struct itimerval));
    setitimer(ITIMER_PROF ,&itval,&itoval);
#endif
    __asm__ __volatile__ ("fsave (%0)\n\t": :"r"(&qw));    
    STORE_ALL;
    result=DriverProc(module->dwDriverID, handle, message, arg1, arg2);
    REST_ALL;
    __asm__ __volatile__ ("fsave (%0)\n\t": :"r"(&qw));    

#ifdef __FreeBSD__
    setitimer(ITIMER_PROF,&itoval,NULL);
#endif
    return result;
}

VideoCodec::VideoCodec(Module* module, int compressor, Module::Mode mode)
    :_module(module)
{
    _handle=_module->CreateHandle(compressor, mode);
}
VideoCodec::~VideoCodec()
{
//    cerr<<"Deleting codec"<<endl;
    _module->CloseHandle(_handle);
}    

VideoCodec* VideoCodecControl::Create(int compressor, const CodecInfo& info, Module::Mode mode)
{
//    cerr<<"VideoCodecControl::Create()"<<endl;
//    cerr<<"_modules size: "<<_modules.size()<<endl;
    list<Module*>::iterator it;
    for(it=_modules.begin(); it!=_modules.end(); it++)
    {
	if((*it)->Name()==info.dll)
	    return new VideoCodec(*it, compressor, mode);
    }
    Module* module=new Module(info.dll, *this);
    _modules.push_back(module);
    return new VideoCodec(module, compressor, mode);
}
VideoCodecControl::~VideoCodecControl()
{
    list<Module*>::iterator it;
    for(it=_modules.begin(); it!=_modules.end(); it++)
    {
	(*it)->ForgetParent();
	delete *it;
    }	
}
void VideoCodecControl::Erase(Module* mod)
{
//    cerr<<"Erasing module "<<(int)mod<<endl;
    list<Module*>::iterator it;
    for(it=_modules.begin(); it!=_modules.end(); it++)
	if(*it==mod)
	{
	    _modules.erase(it);
//	    cerr<<"Success: new _modules size "<<_modules.size()<<endl;
	    return;
	}    
}	    	    
//
//
//  Here goes VFW API
//
//

HRESULT VideoCodec::GetFormat(BITMAPINFOHEADER* v1,BITMAPINFOHEADER* v2)
{
    return _module->Message(_handle, ICM_COMPRESS_GET_FORMAT, (int)v1, (int)v2);
}

HRESULT VideoCodec::GetDefaultKeyFrameRate(int* lpint) 		
{
	return _module->Message(_handle, ICM_GETDEFAULTKEYFRAMERATE,		
	    (long)(void*)(lpint), 0);
}
HRESULT VideoCodec::GetDefaultQuality(int* lpint) 			
{
	return _module->Message(_handle, ICM_GETDEFAULTQUALITY,
	     (long)(void*)(lpint), 0);
}	    	

HRESULT VideoCodec::CompressBegin( BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER* lpbiOutput) 			
{
    return _module->Message(_handle, ICM_COMPRESS_BEGIN,
     (long)(void*)(lpbiInput), (long)(void*)(lpbiOutput));
}    

HRESULT VideoCodec::CompressGetSize( BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER* lpbiOutput) 		
{
    return _module->Message(_handle,							
    	 ICM_COMPRESS_GET_SIZE, (long)(void*)(lpbiInput), 	
	(long)(void*)(lpbiOutput));
}
HRESULT VideoCodec::CompressQuery( BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER* lpbiOutput)
{
    return _module->Message(_handle,						
    	 ICM_COMPRESS_QUERY, (long)(void*)(lpbiInput),
	(long)(void*)(lpbiOutput));
}
HRESULT VideoCodec::CompressEnd()
{
    return _module->Message(_handle, ICM_COMPRESS_END, 0, 0);
}    

HRESULT VideoCodec::Compress(long dwFlags,LPBITMAPINFOHEADER lpbiOutput,void* lpData,
	LPBITMAPINFOHEADER lpbiInput,void* lpBits,long* lpckid,
	long* lpdwFlags,long lFrameNum,long dwFrameSize,long dwQuality,
	LPBITMAPINFOHEADER lpbiPrev,void* lpPrev)
{
	ICCOMPRESS	iccmp;
	iccmp.dwFlags		= dwFlags;
	iccmp.lpbiOutput	= lpbiOutput;
	iccmp.lpOutput		= lpData;
	iccmp.lpbiInput		= lpbiInput;
	iccmp.lpInput		= lpBits;

	iccmp.lpckid		= lpckid;
	iccmp.lpdwFlags		= lpdwFlags;
	iccmp.lFrameNum		= lFrameNum;
	iccmp.dwFrameSize	= dwFrameSize;
	iccmp.dwQuality		= dwQuality;
	iccmp.lpbiPrev		= lpbiPrev;
	iccmp.lpPrev		= lpPrev;
	return _module->Message(_handle,ICM_COMPRESS,(long)&iccmp,sizeof(iccmp));
}

HRESULT VideoCodec::DecompressBegin(BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER*  lpbiOutput) 
{
    return _module->Message(_handle, ICM_DECOMPRESS_BEGIN, (long)(void*)(lpbiInput),	
	(long)(lpbiOutput)				
    );
}
HRESULT VideoCodec::DecompressQuery(BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER* lpbiOutput) 	
{
    return _module->Message(_handle,ICM_DECOMPRESS_QUERY, (long)(void*)(lpbiInput),	
	(long)(lpbiOutput)				
    );
}
HRESULT VideoCodec::DecompressGetFormat(BITMAPINFOHEADER* lpbiInput, BITMAPINFOHEADER* lpbiOutput)
{
    return _module->Message(_handle,ICM_DECOMPRESS_GET_FORMAT, (long)(void*)(lpbiInput),	
	(long)(lpbiOutput));
}

HRESULT VideoCodec::DecompressEnd()
{
     return _module->Message(_handle, ICM_DECOMPRESS_END, 0, 0);
}     
HRESULT VideoCodec::Decompress(long dwFlags,LPBITMAPINFOHEADER lpbiFormat,void* lpData,LPBITMAPINFOHEADER  lpbi,void* lpBits) 
{
	ICDECOMPRESS	icd;
	int result;
	icd.dwFlags	= dwFlags;
	icd.lpbiInput	= lpbiFormat;
	icd.lpInput	= lpData;

	icd.lpbiOutput	= lpbi;
	icd.lpOutput	= lpBits;
	icd.ckid	= 0;
	STORE_ALL;
	result=_module->Message(_handle,ICM_DECOMPRESS,(long)&icd,sizeof(icd));
	REST_ALL;
	return result;
}
