/********************************************************

	Video decoder implementation
	Copyright 2000 Eugene Kuznetsov  (divx@euro.ru)

*********************************************************/

#include <config.h>

#include "VideoDecoder.h"
#include "VideoEncoder.h"
#include "DirectShow/DS_VideoDecoder.h"
#include "Unc.h"
//#include "mjpeg/mjpeg_decoder.h"
#include <string.h>
#include <stdlib.h>
#include <except.h>
#define __MODULE__ "VideoDecoder"

#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdio.h>
#include <unistd.h>
#include <strstream>
#include <dlfcn.h>

#include <registry.h>
#include <wine/winreg.h>

class DlOpener
{
    void* handle;
public:
    DlOpener(){handle=0;}
    ~DlOpener(){if(handle)dlclose(handle);}
    void* open(string name){return handle=dlopen(name.c_str(), RTLD_LAZY);}
    void* sym(string name){return dlsym(handle, name.c_str());}
    void* sym(const char* name){return dlsym(handle, name);}
    void clear(){handle=0;}
};

IVideoDecoder* IVideoDecoder::Create(const BITMAPINFOHEADER& bh, int depth=24, int flip=0)
{
    vector<CodecInfo>::iterator it;
    vector<int>::iterator iv;
    for(it=video_codecs.begin(); it!=video_codecs.end(); it++)
    {
	DlOpener handle;
	IVideoDecoder* (*CreateDecoder)(const BITMAPINFOHEADER&, int);
	for(iv=it->fourcc_array.begin(); iv!=it->fourcc_array.end(); iv++)
	    if(bh.biCompression==*iv)
	    {
		switch(it->kind)
		{
		case CodecInfo::DShow_Dec:
		try
		{
		    return new DS_VideoDecoder(*it, bh, flip);
		}
		catch(FatalError& e)
		{
		    e.Print();
		    continue;
		}    
		case CodecInfo::Win32:
		try
		{
		    return new VideoDecoder(*it, bh, flip);
		}
		catch(FatalError& e)
		{
		    e.Print();
		    continue;
		}    
    		case CodecInfo::Source:
		    return new Unc_Decoder(*it, bh);
		    break;
		case CodecInfo::Plugin:
		try
		{
		    if(!handle.open(it->dll))continue;
		    CreateDecoder=
	(IVideoDecoder* (*)(const BITMAPINFOHEADER&, int))handle.sym("CreateDecoder");
		    if(!CreateDecoder)
			continue;
		    handle.clear();
		    return CreateDecoder(bh, flip);
		    continue;
		}
		catch(FatalError& e)
		{
		    e.Print();
		    continue;
		}   
    
		}    
	    }	
    }
    strstream err;
    err.setf(ios::hex, ios::basefield);
    char s[5];
    *(int*)s=bh.biCompression;
    s[4]=0;
    err<<"Unknown codec "<<bh.biCompression<<" = '"<<s<<"'!"<<endl;
    throw FATAL(err.str());
}

VideoDecoder::VideoDecoder(const CodecInfo& info, const BITMAPINFOHEADER& format, int flip)
    :m_iState(0), field_num(0), codec(0), divx_trick(0), record(info)
{
    m_outFrame=0;
    try
    {
	m_bh=format;
	memset(&m_obh, 0, sizeof(m_obh));
	m_obh.biSize=sizeof(m_obh);
#warning NOT SURE
//
// MPEG-4 format has changed during development. Package
// binaries.zip includes mpg4c32.dll - old MPEG-4 codec
// with unlocked compression - and divxc32.dll,
// which differs from current MPEG-4 only by FOURCC's.
// The problem is to identify which is needed to us,
// because 'old' codec doesn't work with 'new' clips and
// vice versa. Unfortunately, they may simulate correct
// work until the moment of actual decompression. Then, 
// you'll receive garbled picture and lots of warnings about
// hr=-100.
//

// I've got old video with fccHandler mp43 and biCompression MPG4.
// Dirk Vornheider sent me old video with both values MP42.

	switch(m_bh.biCompression)
	{
	    case mmioFOURCC('M', 'P', '4', '1'):
	    case mmioFOURCC('M', 'P', '4', '3'):
		m_bh.biCompression=mmioFOURCC('d', 'i', 'v', '3');
	        break;
	}	    	
	codec=control.Create(m_bh.biCompression, info, Module::Decompress);

	if(codec==0)
	    throw FATAL("ICOpen failed");
    
        HRESULT h=codec->DecompressGetFormat(&m_bh, (BITMAPINFOHEADER*)&m_obh);
        if(h!=0)
	{
	    printf("Declined format dump:\n");
    	    printf("  biSize %d\n", m_bh.biSize);
	    printf("  biWidth %d\n", m_bh.biWidth);
	    printf("  biHeight %d\n", m_bh.biHeight);
	    printf("  biPlanes %d\n", m_bh.biPlanes);
	    printf("  biBitCount %d\n", m_bh.biBitCount);
	    printf("  biCompression %d='%.4s'\n", m_bh.biCompression, &m_bh.biCompression);
	    printf("  biSizeImage %d\n", m_bh.biSizeImage);
	    printf(" Error %d\n", h);
	    throw FATAL("ICDecompressGetFormat failed");
	}

	if(!flip)
	{
	    m_obh.biHeight*=-1;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&m_obh);
	    if(h!=0)
		throw FATAL("Decoder does not support upside-down frames");
	}	
	BitmapInfo mbi(m_obh);
	mbi.setBitFields16();
	if(info.fourcc_array[0]==mmioFOURCC('M', 'J', 'P', 'G'))
	    mbi.biSize=0x28;
	h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
        if(h!=0)
	    throw FATAL("Decoder is not capable of 16-bit(565) output");
	mbi.setRGB();
	caps=(CAPS)0;	
	int caps_test=1;	
	switch(info.fourcc_array[0])
	{
	    case mmioFOURCC('D', 'I', 'V', '3'):
	    case mmioFOURCC('D', 'I', 'V', '4'):
	    case mmioFOURCC('M', 'P', '4', '2'):
		divx_trick=1;
		caps_test=0;
		caps=CAP_YUY2; //planar formats broken, others untested
		break;
	    case mmioFOURCC('I', 'V', '5', '0'):
		caps_test=0;
		caps=CAP_YUY2; //planar formats broken, others untested
		break;
	}
	if(caps_test)
	{
	    mbi.biCompression=fccYUY2;
	    mbi.biBitCount=16;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
	    if(!h) caps=(CAPS)(caps | CAP_YUY2);

    	    mbi.biCompression=fccIYUV;
	    mbi.biBitCount=12;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
	    if(!h) caps=(CAPS)(caps | CAP_IYUV);

	    mbi.biCompression=fccUYVY;
	    mbi.biBitCount=16;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
	    if(!h) caps=(CAPS)(caps | CAP_UYVY);

	    mbi.biCompression=fccYV12;
	    mbi.biBitCount=12;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
	    if(!h) caps=(CAPS)(caps | CAP_YV12);

	    mbi.biCompression=fccYVYU;
	    mbi.biBitCount=16;
	    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&mbi);
	    if(!h) caps=(CAPS)(caps | CAP_YVYU);
	}
	if(caps)
	{
	    cerr<<"Decoder is capable of YUV output ( flags 0x"<<hex<<(int)caps<<dec<<" )"<<endl;
	}
        m_iState=1;		
	return;
    }
    catch(FatalError& error)
    {
        delete codec;
        delete m_outFrame;
	m_outFrame=0;
	codec=0;
	throw;	
    }	    
}
void VideoDecoder::Start()
{
    if(m_iState!=1)
    return;
/*
      printf("Starting decompression, format: \n");
	printf("  biSize %d\n", m_bh.biSize);
	printf("  biWidth %d\n", m_bh.biWidth);
	printf("  biHeight %d\n", m_bh.biHeight);
	printf("  biPlanes %d\n", m_bh.biPlanes);
	printf("  biBitCount %d\n", m_bh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_bh.biCompression, &m_bh.biCompression);
	printf("  biSizeImage %d\n", m_bh.biSizeImage);
    printf("Dest fmt:\n");
	printf("  biSize %d\n", m_obh.biSize);
	printf("  biWidth %d\n", m_obh.biWidth);
	printf("  biHeight %d\n", m_obh.biHeight);
	printf("  biPlanes %d\n", m_obh.biPlanes);
	printf("  biBitCount %d\n", m_obh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_obh.biCompression, &m_obh.biCompression);
	printf("  biSizeImage %d\n", m_obh.biSizeImage);
*/
    int tmp;
    if(divx_trick){tmp=m_obh.biCompression; m_obh.biCompression=0;}
    HRESULT hr=codec->DecompressBegin(&m_bh, (BITMAPINFOHEADER*)&m_obh);
    if(divx_trick)m_obh.biCompression=tmp;
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressBegin() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
    m_outFrame=new CImage(&m_obh);
    m_iState=2;
    return;		
}
void VideoDecoder::Stop()
{
    if(m_iState!=2)
	return;
    HRESULT hr=codec->DecompressEnd();
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressEnd() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
    m_outFrame->release();//just in case
    m_outFrame=0;
    m_iState=1;
    return;
}
void VideoDecoder::Restart()
{
    if(m_iState!=2)
	return;
    HRESULT hr=codec->DecompressEnd();
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressEnd() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
    int tmp;
    if(divx_trick){tmp=m_obh.biCompression; m_obh.biCompression=0;}
    hr=codec->DecompressBegin(&m_bh, (BITMAPINFOHEADER*)&m_obh);
    if(divx_trick)m_obh.biCompression=tmp;
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressBegin() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
}
void VideoDecoder::Close()
{
    if(m_iState==0)
	return;
    delete codec;
    delete m_outFrame;
    codec=0;
    m_outFrame=0;
    m_iState=0;
}
int VideoDecoder::DecodeFrame(char* src, int size, int is_keyframe)
{
	m_bh.biSizeImage=size;
//	cerr<<"DecodeFrame called"<<endl;
	if(size)
	{
	    int hr=codec->Decompress((is_keyframe?0:ICDECOMPRESS_NOTKEYFRAME),			
		 &m_bh, src, (BITMAPINFOHEADER*)m_outFrame->get_fmt(), m_outFrame->data());
//	cerr<<"Result "<<hr<<endl;
	    if(hr!=0)cerr<<"VideoDecoder: warning: hr="<<hr<<endl;
	    return hr;
	}    
	return 0;
}
int VideoDecoder::SetDestFmt(int bits=24, int csp=0)
{
    if(m_iState==0)
	return -1;
    if(CImage::UnknownColorSpace(csp))
	return -1;
    HRESULT h;
    BitmapInfo temp=m_obh;
    if(csp==0)
    {
	temp.biBitCount=bits;
        temp.biSizeImage=abs(temp.biWidth*temp.biHeight*temp.biBitCount/8);
	switch(bits)
        {
	case 15:
	    temp.setBitFields15();
    	    break;	    
	case 16:
	    temp.setBitFields16();
	    if(record.fourcc_array[0]==mmioFOURCC('M', 'J', 'P', 'G'))
		temp.biSize=0x28;
	    break;	    
        default:
    	    temp.setRGB();
	    break;
        }
    }	
    else
    {
	temp.setRGB();
	temp.biBitCount=CImage::BitCount(csp);
	temp.biCompression=csp;
        temp.biSizeImage=abs(temp.biWidth*temp.biHeight*CImage::BitCount(csp)/8);
    }
    int tmp;
    if(divx_trick && csp && (csp!=3))
	{tmp=temp.biCompression; temp.biCompression=0;}
    h=codec->DecompressQuery(&m_bh, (BITMAPINFOHEADER*)&temp);
    if(divx_trick && csp && (csp!=3))
	temp.biCompression=tmp;
    
    if(h!=0)
    {
	if(csp)
	    cerr<<"Unsupported color space"<<endl;
	else
	    cerr<<"Unsupported bit depth"<<endl;
	return -1;
    }    
//    delete m_outFrame;
//    m_outFrame=0;
//    m_outFrame=new CImage(&temp);
    m_obh=temp;
    if(csp)
	m_obh.biBitCount=CImage::BitCount(csp);
    m_bh.biBitCount=bits;
    if(m_iState==2)
    {
        Stop();
        Start();
    }
    return 0; 
}
int Unc_Decoder::DecodeFrame(char* src, int size, int is_keyframe)
{
    if(!src)return -100;
    if(size<m_outFrame->bytes())return -100;
    memcpy(m_outFrame->data(), src, size);
    return 0;
}
Unc_Decoder::Unc_Decoder(const CodecInfo& info, const BITMAPINFOHEADER& h):record(info)
{
    m_outFrame=0;
    m_bh=h;
    m_obh=h;
    m_obh.biHeight*=-1;
}
Unc_Decoder::~Unc_Decoder()
{
    Stop();
}
void Unc_Decoder::Start()
{
    BitmapInfo i(m_bh);
    delete m_outFrame;
    m_outFrame=new CImage(&i);
}
void Unc_Decoder::Stop()
{
    delete m_outFrame;
    m_outFrame=0;
}

CImage* IVideoDecoder::GetFrame()
{
    if(!m_outFrame)return 0;
    if(m_outFrame->is_fmt(&m_obh))
    {
	m_outFrame->addref();
	return m_outFrame;
    }
    else
	return new CImage(m_outFrame, &m_obh);
}
int IVideoDecoder::SetDirection(int d)
{
    m_obh.biHeight=d?m_bh.biHeight:(-m_bh.biHeight);	
    return 0;
}



int IVideoDecoder::GetExtendedAttr(const CodecInfo& info, const char* attribute, int& value)
{
    int result, newkey, status, count;
    if((info.fourcc==fccDIV3)||(info.fourcc==fccDIV4))
    {
	if(strcmp(attribute, "Quality")==0)
	{
	    char* keyname;
    	    if(info.kind==CodecInfo::Win32)
		keyname="SOFTWARE\\Microcrap\\Scrunch";
	    else
		keyname="SOFTWARE\\Microsoft\\Scrunch";
    	    result=RegCreateKeyExA(HKEY_CURRENT_USER, keyname, 0, 0, 0, 0, 0,	  		&newkey, &status);
            if(result!=0)
	    {
	        printf("VideoDecoder::GetExtAttr: registry failure\n");
	        return -1;
	    }    
	    count=4;
	    result=RegQueryValueExA(newkey, "Current Post Process Mode", 0, &status, &value, &count);
            if(result!=0)
	    {
	        printf("VideoEncoder::GetExtAttr: no such value\n");
	        value=-1;
	        return -1;
	    }    
   	    RegCloseKey(newkey);
   	    return result;
	}
    }
    if(
	(info.fourcc==fccDIV3)||
	(info.fourcc==fccDIV4)||
	(info.fourcc==fccWMV1)||
	(info.fourcc==fccWMV2)
	)
    {
	if(
	(strcmp(attribute, "Saturation")==0) ||
	(strcmp(attribute, "Hue")==0) ||
	(strcmp(attribute, "Contrast")==0) ||
	(strcmp(attribute, "Brightness")==0)
	)
	{
	    char* keyname="SOFTWARE\\Microsoft\\Scrunch\\Video";
    	    result=RegCreateKeyExA(HKEY_CURRENT_USER, keyname, 0, 0, 0, 0, 0,	  		&newkey, &status);
            if(result!=0)
	    {
	        printf("VideoDecoder::GetExtAttr: registry failure\n");
	        return -1;
	    }    
	    count=4;
	    result=RegQueryValueExA(newkey, attribute, 0, &status, &value, &count);
            if(result!=0)
	    {
	        printf("VideoEncoder::GetExtAttr: no such value\n");
	        value=50;
	        return 0;
	    }    
   	    RegCloseKey(newkey);
   	    return result;
	}

    }
}   
int IVideoDecoder::SetExtendedAttr(const CodecInfo& info, const char* attribute, int value)
{
    int result, newkey, status;
    if((info.fourcc==fccDIV3)||(info.fourcc==fccDIV4))
    {
        if(strcmp(attribute, "Quality")==0)
        {
	    char* keyname;
	    if(info.kind==CodecInfo::Win32)
		keyname="SOFTWARE\\Microcrap\\Scrunch";
	    else
		keyname="SOFTWARE\\Microsoft\\Scrunch";
    	    result=RegCreateKeyExA(HKEY_CURRENT_USER, keyname, 0, 0, 0, 0, 0,	   		&newkey, &status);
            if(result!=0)
	    {
	        printf("VideoDecoder::SetExtAttr: registry failure\n");
	        return -1;
	    }    
	    result=RegSetValueExA(newkey, "Current Post Process Mode", 0, REG_DWORD, &value, 4);
            if(result!=0)
	    {
	        printf("VideoDecoder::SetExtAttr: error writing value\n");
	        return -1;
	    }    
	    value=-1;
	    result=RegSetValueExA(newkey, "Force Post Process Mode", 0, REG_DWORD, &value, 4);
            if(result!=0)
	    {
		printf("VideoDecoder::SetExtAttr: error writing value\n");
	    	return -1;
	    }    
   	    RegCloseKey(newkey);
   	    return 0;
	}   	
    }
    if(
	(info.fourcc==fccDIV3)||
	(info.fourcc==fccDIV4)||
	(info.fourcc==fccWMV1)||
	(info.fourcc==fccWMV2)
	)
    {
        if(
	(strcmp(attribute, "Saturation")==0) ||
	(strcmp(attribute, "Hue")==0) ||
	(strcmp(attribute, "Contrast")==0) ||
	(strcmp(attribute, "Brightness")==0)
	)
        {
	    char* keyname="SOFTWARE\\Microsoft\\Scrunch\\Video";
    	    result=RegCreateKeyExA(HKEY_CURRENT_USER, keyname, 0, 0, 0, 0, 0,	   		&newkey, &status);
            if(result!=0)
	    {
	        printf("VideoDecoder::SetExtAttr: registry failure\n");
	        return -1;
	    }    
	    result=RegSetValueExA(newkey, attribute, 0, REG_DWORD, &value, 4);
            if(result!=0)
	    {
	        printf("VideoDecoder::SetExtAttr: error writing value\n");
	        return -1;
	    }    
   	    RegCloseKey(newkey);
   	    return 0;
	}   	
    }	
    return -1;
}
