/********************************************************

	Video encoder implementation
	Copyright 2000 Eugene Kuznetsov (divx@euro.ru)

*********************************************************/

#include <config.h>

#include "Unc.h"
#include "VideoEncoder.h"
//#include "mjpeg/mjpeg_encoder.h"
#include <time.h>
#include <registry.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <iostream>
#include <strstream>
//#include <dirent.h>
#include <dlfcn.h>
using namespace std;

#define HKEY_CLASSES_ROOT       (0x80000000)
#define HKEY_CURRENT_USER       (0x80000001)
#define HKEY_LOCAL_MACHINE      (0x80000002)
#define HKEY_USERS              (0x80000003)
#define HKEY_PERFORMANCE_DATA   (0x80000004)
#define HKEY_CURRENT_CONFIG     (0x80000005)
#define HKEY_DYN_DATA           (0x80000006)
#define REG_DWORD		4	/* DWORD in little endian format */
//#define TIMING
#include <except.h>
//#undef TIMING
#define __MODULE__ "VideoEncoder"
#ifdef TIMING
#include "aviutil.h"
#endif
class DlOpener
{
    void* handle;
public:
    DlOpener(){handle=0;}
    ~DlOpener(){if(handle)dlclose(handle);}
    void* open(string name){return handle=dlopen(name.c_str(), RTLD_LAZY);}
    void* sym(string name){return dlsym(handle, name.c_str());}
    void* sym(const char* name){return dlsym(handle, name);}
    void clear(){handle=0;}
};
IVideoEncoder* IVideoEncoder::Create(int compressor, const BITMAPINFOHEADER& bh)
{
    vector<CodecInfo>::iterator it;
    vector<int>::iterator iv;
    for(it=video_codecs.begin(); it!=video_codecs.end(); it++)
    {
	DlOpener handle;
	IVideoEncoder* (*CreateEncoder)(int compressor, const BITMAPINFOHEADER& bh);	
	for(iv=it->fourcc_array.begin(); iv!=it->fourcc_array.end(); iv++)
	    if(compressor==*iv)
		switch(it->kind)
		{
		case CodecInfo::Win32:
		try
		{
		    return new VideoEncoder(compressor, *it, bh);
		}
		catch(FatalError& e)
		{
		    e.Print();
		    continue;
		}    
		case CodecInfo::Source:
		    return new Unc_Encoder(*it, bh);
		    break;
//		    return new MJPEG_VideoEncoder(compressor, *it, bh);
		case CodecInfo::Plugin:
		try
		{
		    if(!handle.open(it->dll))continue;
		    CreateEncoder=
	(IVideoEncoder* (*)(int, const BITMAPINFOHEADER&))handle.sym("CreateEncoder");
		    if(!CreateEncoder)
			continue;
		    handle.clear();//warning
		    //it won't be unloaded
		    return CreateEncoder(compressor, bh);
		    continue;
		}    
		catch(FatalError& e)
		{
		    e.Print();
		    continue;
		}    
		}    
    }

    strstream err;
    err.setf(ios::hex, ios::basefield);
    char s[5];
    *(int*)s=compressor;
    s[4]=0;
    err<<"Unknown codec "<<compressor<<" = '"<<s<<"'!"<<endl;
    throw FATAL(err.str());
}
IVideoEncoder* IVideoEncoder::Create(const VideoEncoderInfo& info)
{
    IVideoEncoder* en=Create(info.compressor, info.header);
    en->SetQuality(info.quality);
    en->SetKeyFrame(info.keyfreq);
    return en;
}

int IVideoEncoder::SetRegValue(int fccHandler, const char* name, int value)
{
   int result, status, newkey;
   if(name==0)
   	return -1;
   char full_name[50]="Software\\LinuxLoader\\"; 
   char hn[5];
   hn[0]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[1]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[2]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[3]=tolower(fccHandler & 0xFF);
   hn[4]=0;
   strcat(full_name, hn);
   result=RegCreateKeyExA(HKEY_CURRENT_USER, full_name, 0, 0, 0, 0, 0,
    &newkey, &status);
   if(result!=0)
    return -1;
   result=RegSetValueExA(newkey, name, 0, REG_DWORD, &value, 4);
   RegCloseKey(newkey);
   return result;
}
int IVideoEncoder::GetRegValue(int fccHandler, const char* name, int* place)
{
   int result, status, newkey, count;
   if(name==0)
   	return -1;
   if(place==0)
   	return -1;
   char full_name[50]="Software\\LinuxLoader\\"; 
   char hn[5];
   hn[0]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[1]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[2]=tolower(fccHandler & 0xFF);
   fccHandler>>=8;
   hn[3]=tolower(fccHandler & 0xFF);
   hn[4]=0;
   strcat(full_name, hn);
   result=RegOpenKeyExA(HKEY_CURRENT_USER, full_name, 0, 0, &newkey);
   if(result!=0)
    return -1;
   count=4;
   result=RegQueryValueExA(newkey, name, 0, 0, &status, &count);
   if(count!=4)result=-1;
   RegCloseKey(newkey);
   if(result==0)*place=status;
   return result;
}

VideoEncoder::VideoEncoder(int compressor, const CodecInfo& info, const BITMAPINFOHEADER& format)
    :m_iState(0), m_prev(0), m_bh(0), m_obh(0), codec(0), record(info)
//void VideoEncoder::Init(int compressor, const char* format) throw (FatalError&)
{
    try
    {
	if(m_iState>0)
	    return;//already initialized
//	m_bh=(BITMAPINFOHEADER*)(new char[sizeof(*m_bh)]);
//	*m_bh=format;
	m_bh=new BitmapInfo(format);
	
	switch(compressor)
	{
	    case mmioFOURCC('M', 'P', '4', '1'):
	    case mmioFOURCC('M', 'P', '4', '3'):
		compressor=mmioFOURCC('d', 'i', 'v', '3');
	        break;
	}	    	
	codec=control.Create(compressor, info, Module::Compress);
	m_comp_id=compressor;
	int h=codec->GetFormat(m_bh, 0);
	if(h<0)
	    throw FATAL("Can't handle this format");
	m_obh=(BITMAPINFOHEADER*)new char[h];
	memset(m_obh, 0, h);
	m_obh->biSize=h;
	h=codec->GetFormat(m_bh, m_obh);
	m_obh->biHeight=abs(m_obh->biHeight);
	m_obh->biBitCount=m_bh->biBitCount;
        if(h!=0)
	    throw FATAL("Can't handle this format");
//	m_obh->biHeight*=-1;
//        ICGetDefaultQuality(hic, &m_iQual);
	codec->GetDefaultQuality(&m_iQual);
	m_iState=1;		
//        if(ICGetDefaultKeyFrameRate(hic, &m_iKfFreq)!=0)
	if(codec->GetDefaultKeyFrameRate(&m_iKfFreq)!=0)
 	    m_iKfFreq=0xFFFF;
	if(m_bh->biSizeImage==0)
    	    m_bh->biSizeImage=abs(m_bh->biWidth*m_bh->biHeight*((m_bh->biBitCount+7)/8));
	return;
    }
    catch(...)
    {
	if(codec)delete codec;
//	if(hic)ICClose(hic);
	throw;
    }	
}
void VideoEncoder::Start()
{
    if(m_iState!=1)
	return;//wrong state
//    HRESULT hr=ICCompressBegin(hic, m_bh, m_obh);
    HRESULT hr=codec->CompressBegin(m_bh, m_obh);
/*	printf("Starting compression \n");
	printf("  biSize %d\n", m_bh->biSize);
	printf("  biWidth %d\n", m_bh->biWidth);
	printf("  biHeight %d\n", m_bh->biHeight);
	printf("  biPlanes %d\n", m_bh->biPlanes);
	printf("  biBitCount %d\n", m_bh->biBitCount);
	printf("  biCompression %d='%.4s'\n", m_bh->biCompression, &m_bh->biCompression);
	printf("  biSizeImage %d\n", m_bh->biSizeImage);
	printf("Dest fmt:\n");	
	printf("  biSize %d\n", m_obh->biSize);
	printf("  biWidth %d\n", m_obh->biWidth);
	printf("  biHeight %d\n", m_obh->biHeight);
	printf("  biPlanes %d\n", m_obh->biPlanes);
	printf("  biBitCount %d\n", m_obh->biBitCount);
	printf("  biCompression %d='%.4s'\n", m_obh->biCompression, &m_obh->biCompression);
	printf("  biSizeImage %d\n", m_obh->biSizeImage);
*/
    if(hr!=0)
	cerr<<"ICCompressBegin() failed ( shouldn't happen ), error code "<<hr<<endl;
    m_iFrameNum=0;
    m_iState=2;
    return;		
}
void VideoEncoder::Stop()
{
    if(m_iState!=2)
	return;//wrong state
//    HRESULT hr=ICCompressEnd(hic);
    HRESULT hr=codec->CompressEnd();
    if(hr!=0)
	cerr<<"ICCompressEnd() failed ( shouldn't happen ), error code "<<hr<<endl;
    m_iState=1;
    return;		
}
void VideoEncoder::Close()
{
    if(m_iState==0)
	return;
//    ICClose(hic);
    delete codec;
    if(m_prev)delete m_prev;
    m_iState=0;
    return;
}
int VideoEncoder::QueryOutputSize()
{
    if(m_iState==0)
	return -1;
	
//    return ICCompressGetSize(hic, m_bh, m_obh);    
    return codec->CompressGetSize(m_bh, m_obh);    
}
int VideoEncoder::EncodeFrame(CImage* src, char* dest, int* is_keyframe, int* size, int* lpckid)
{
	int st1, st2;
        static BITMAPINFOHEADER m_prevbh;
        if(m_iState!=2)
	    return -1;
	if(!src)return -1;
#ifdef TIMING
	st1=localcount();    
#endif	
	CImage* temp=0;
//	if(src->depth()!=m_bh->biBitCount)
//	    temp=new CImage(src, m_bh->biBitCount);
	if(!src->is_fmt(m_bh))
	    temp=new CImage(src, m_bh);
//	int result=ICCompress(hic, (m_iKfFreq ? ((m_iFrameNum%m_iKfFreq)? 0 : ICCOMPRESS_KEYFRAME) : 0),
//	  m_obh, dest, m_bh, src,
//	  (long*)lpckid, (long*)is_keyframe, m_iFrameNum, 0, 
//	  m_iQual, &m_prevbh, m_prev);
	int result=codec->Compress((m_iKfFreq ? ((m_iFrameNum%m_iKfFreq)? 0 : ICCOMPRESS_KEYFRAME) : 0),
	  m_obh, dest, m_bh, temp?temp->data():src->data(),
	  (long*)lpckid, (long*)is_keyframe, m_iFrameNum, 0, 
	  m_iQual, &m_prevbh, m_prev);
	if(temp)temp->release();
	if(result==0)
	{
	    //if(m_prev==0)m_prev=new char[m_bh->biSizeImage];
	    //memcpy(m_prev, src, m_bh->biSizeImage);
	    if(m_prev==0)m_prev=new char[codec->CompressGetSize(m_bh, m_obh)];
	    memcpy(m_prev, dest, m_obh->biSizeImage);
            m_prevbh=*m_bh;
	}
#ifdef TIMING
	st2=localcount();
	printf("ICCompress: %f ms into %d bytes\n", (double)(st2-st1)/freq, m_obh->biSizeImage);
#endif
	m_iFrameNum++;
	*size=m_obh->biSizeImage;
	return result;
}
int VideoEncoder::SetQuality(int quality)
{
    if(m_iState==0)
	return -1;
    if(quality<0)
	return -1;	
    if(quality>10000)
	return -1;
    m_iQual=quality;	
    return 0;
}
int VideoEncoder::GetQuality()
{
    return m_iQual;
}

int VideoEncoder::SetKeyFrame(int freq)
{
    if(m_iState==0)
	return -1;
    if(freq<=0)	
	codec->GetDefaultKeyFrameRate(&m_iKfFreq);
//	ICGetDefaultKeyFrameRate(hic, &m_iKfFreq);
    else
	m_iKfFreq=freq;	
    return 0;	
}

int IVideoEncoder::SetExtendedAttr(int comp_id, const char* attribute, int value)
{
   int result, status, newkey, count;
    
    if(attribute==0)
	return -1;
    switch(comp_id)
    {
    case fccIV50:
        
 	if(strcmp(attribute, "QuickCompress")==0)
     	{
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\Intel\\Indeo\\5.0", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure\n");
		    return -1;
		}    
		result=RegSetValueExA(newkey, "QuickCompress", 0, REG_DWORD, &value, 4);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
   		return result;
	}   	
    
    	printf("Unknown attribute '%s' for Intel Indeo 5.0 compressor\n", attribute);
    break;

    case fccIV41:
        
 	if(strcmp(attribute, "QuickCompress")==0)
     	{
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\Intel\\Indeo\\4.1", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)return -1;
		result=RegSetValueExA(newkey, "QuickCompress", 0, REG_DWORD, &value, 4);
   		RegCloseKey(newkey);
   		return result;
	}   	
    
    	printf("Unknown attribute '%s' for Intel Indeo 4.1 compressor\n", attribute);
    break;
    
    case fccDIV3:
    case fccDIV4:
    if(
     (strcmp(attribute, "Crispness")==0) || (strcmp(attribute, "KeyFrames")==0)
       ) 
       return SetRegValue(comp_id, attribute, value);
    if(strcmp(attribute, "BitRate")==0)
     	{
		SetRegValue(comp_id, attribute, value);
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "SOFTWARE\\Microcrap\\Scrunch\\Video", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure\n");
		    return -1;
		}    
		result=RegSetValueExA(newkey, "BitRate", 0, REG_DWORD, &value, 4);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\LinuxLoader\\Div3", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure\n");
		    return -1;
		}    
		result=RegSetValueExA(newkey, "BitRate", 0, REG_DWORD, &value, 4);
           	if(result!=0)
		{
		    printf("VideoEncoder::SetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
   		return result;
	}   	
    
    
    printf("Unknown attribute '%s' for DivX compressor\n", attribute);
    break;
    case fccMJPG:
	if(strcmp(attribute, "Mode")==0)
	{
	    char s[256];
	    sprintf(s, "%d", value);
	    _WritePrivateProfileStringA("Compress", "Mode", s, "M3JPEG.INI");
	    return 0;
	}
	printf("Unknown attribute '%s' for Motion JPEG compressor\n", attribute);
	break;
    }
    return -1;
}
int IVideoEncoder::GetExtendedAttr(int comp_id, const char* attribute, int &value)
{
    int result, status, newkey, count;

    if(attribute==0)
	return -1;
    switch(comp_id)
    {
    case fccIV50:
        
 	if(strcmp(attribute, "QuickCompress")==0)
     	{
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\Intel\\Indeo\\5.0", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure\n");
		    return -1;
		}    
		count=4;
		result=RegQueryValueExA(newkey, "QuickCompress", 0, &status, &value, &count);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
   		return result;
	}   	
    
    	printf("Unknown attribute '%s' for Intel Indeo 5.0 compressor\n");
    break;

    case fccIV41:
        
 	if(strcmp(attribute, "QuickCompress")==0)
     	{
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\Intel\\Indeo\\4.1", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure\n");
		    return -1;
		}    
		count=4;
		result=RegQueryValueExA(newkey, "QuickCompress", 0, &status, &value, &count);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
   		return result;
	}   	
    
    	printf("Unknown attribute '%s' for Intel Indeo 4.1 compressor\n");
    break;
    
    case fccDIV3:
    case fccDIV4:
    if(
    (strcmp(attribute, "Crispness")==0) || (strcmp(attribute, "KeyFrames")==0)
       ) 
       return GetRegValue(comp_id, attribute, &value);
        if(strcmp(attribute, "BitRate")==0)
     	{
		result=RegCreateKeyExA(HKEY_CURRENT_USER, "SOFTWARE\\Microcrap\\Scrunch\\Video", 0, 0, 0, 0, 0,
	   		&newkey, &status);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure\n");
		    return -1;
		}    
		count=4;
		result=RegQueryValueExA(newkey, "BitRate", 0, &status, &value, &count);
           	if(result!=0)
		{
		    printf("VideoEncoder::GetExtAttr: registry failure 2\n");
		    return -1;
		}    
   		RegCloseKey(newkey);
   		return result;
	}   	
    
	
    printf("Unknown attribute '%s' for DivX compressor\n", attribute);
    break;
    

    case fccMJPG:
	if(strcmp(attribute, "Mode")==0)
	{
	    value=_GetPrivateProfileIntA("Compress", "Mode", 1, "M3JPEG.INI");
	    return 0;
	}
	printf("Unknown attribute '%s' for Motion JPEG compressor\n", attribute);
	break;
    }
    return -1;
}

int IVideoEncoder::SetExtendedAttr(int comp_id, const char* attribute, const char* value)
{
    if(attribute==0)
	return -1;
    switch(comp_id)
    {
	case fccMJPG:
	    if(!strcmp(attribute, "UserName"))
	    {
		_WritePrivateProfileStringA("Register", "UserName", value, "M3JPEG.INI");
		return 0;		
	    }
	    if(!strcmp(attribute, "LicenseKey"))
	    {
		_WritePrivateProfileStringA("Register", "LicenseKeyV2", value, "M3JPEG.INI");
		return 0;		
	    }
	break;
    }    
    return -1;
}    
int IVideoEncoder::GetExtendedAttr(int comp_id, const char* attribute, char* value, int size)
{
    if(attribute==0)
	return -1;
    switch(comp_id)
    {
	case fccMJPG:
	    if(!strcmp(attribute, "UserName")) 
		return _GetPrivateProfileStringA("Register", "UserName", "", value, size, "M3JPEG.INI");
	    if(!strcmp(attribute, "LicenseKey"))
		return _GetPrivateProfileStringA("Register", "LicenseKeyV2", "", value, size, "M3JPEG.INI");
	break;
    }    
    return -1;
}    
        


VideoEncoder::~VideoEncoder()
{
    if(m_iState>0)
	Close();
    if(m_bh)delete m_bh;
    if(m_obh)delete m_obh;
}    	
const BITMAPINFOHEADER& VideoEncoder::QueryOutputFormat() const
{
    return *m_obh;
}	
