#include "VideoEncoder.h"
#include <dlfcn.h>
#include <dirent.h>
#include "DirectShow/guids.h"
vector<CodecInfo> video_codecs;
map<int, vector<AttributeInfo> > video_attributes;
map<int, vector<AttributeInfo> > decoder_attributes;
extern "C" char* def_path;

GUID wmv1_clsid={0x4facbba1, 0xffd8, 0x4cd7,
    {0x82, 0x28, 0x61, 0xe2, 0xf6, 0x5c, 0xb1, 0xae}};
GUID wmv2_clsid={0x521fb373, 0x7654, 0x49f2, 
    {0xbd, 0xb1, 0x0c, 0x6e, 0x66, 0x60, 0x71, 0x4f}};    
GUID dvsd_clsid={0xB1B77C00, 0xC3E4, 0x11CF,
    {0xAF, 0x79, 0x00, 0xAA, 0x00, 0xB6, 0x7A, 0x42}};
//GUID wmv_cust_iid={0xDA7CC2B7, 0xA246, 0x48D3,
//    {0xB2, 0x5C, 0xd6, 0xb0, 0xde, 0x60, 0xc4, 0x9e}};
CodecInfoKeeper codec_keeper;
//do not rearrange fourcc's in arrays
CodecInfoKeeper::CodecInfoKeeper()
{
    int div3_codecs[]=
    {
	fccDIV3,
	fccdiv3,
	fccMP41,
	fccMP43,
	0
    };
    int div4_codecs[]=
    {
	fccDIV4,
        fccdiv4,
	0
    };
    int iv50_codecs[]=
    {
	mmioFOURCC('I', 'V', '5', '0'),
	mmioFOURCC('i', 'v', '5', '0'),
	0
    };
    int iv41_codecs[]=
    {
	mmioFOURCC('I', 'V', '4', '1'),
	mmioFOURCC('i', 'v', '4', '1'),
	0
    };
    int iv32_codecs[]=
    {
	mmioFOURCC('I', 'V', '3', '2'),
	mmioFOURCC('i', 'v', '3', '2'),
	0
    };
    int iv31_codecs[]=
    {    
	mmioFOURCC('I', 'V', '3', '1'),
	mmioFOURCC('i', 'v', '3', '1'),
	0
    };  
    int mpg4_codecs[]=
    {  
	fccMP42,
        fccmp42,
        fccmp43,
        fccmpg4,
        fccMPG4,
	0
    };
    int cvid_codecs[]=
    {
	mmioFOURCC('c', 'v', 'i', 'd'),
	0
    };
    int vcr2_codecs[]=
    {
	mmioFOURCC('V', 'C', 'R', '2'),
	0
    };
    int i263_codecs[]=
    {
	mmioFOURCC('i', '2', '6', '3'),
	mmioFOURCC('I', '2', '6', '3'),
	0
    };	
    int mjpg_codecs[]=
    {
	mmioFOURCC('M', 'J', 'P', 'G'),//fccMJPG
	mmioFOURCC('m', 'j', 'p', 'g'),
	0
    };
    int dvsd_codecs[]=
    {
	mmioFOURCC('d', 'v', 's', 'd'),
	0
    };
    int apxx_codecs[]=
    {
	mmioFOURCC('A', 'P', '4', '1'),
	mmioFOURCC('a', 'p', '4', '1'),
	mmioFOURCC('A', 'P', '4', '2'),
	mmioFOURCC('a', 'p', '4', '2'),
	0
    };
    int wmv1_codecs[]=
    {
	fccWMV1,
	fccwmv1,
	0
    };
    int wmv2_codecs[]= /* it may be able to decode wmv1, too */
    {
	fccWMV2,
	fccwmv2,
	0
    };
    int zero[]={0};
    char divx_about[] = 
    "Very popular MPEG-4 video codec, based on "
    "Microsoft MPEG-4 implementation. Advanced "
    "compression technologies allow it to "
    "compress 640x480x25 video with a perfect"
    " quality into 100-150 kbytes/s ( 3-4 times"
    " less than MPEG-2 ), by the cost of increased"
    " requirements for playback. \n\n"
    "Web site: <a href=\"http://divx.ctw.cc\">http://divx.ctw.cc</a>";
    char ivxx_about[] = 
    "A set of wavelet video codecs, developed "
    "by Intel and currently owned by Ligos corp."
    " Indeo Video 5.0 with turned on \"QuickCompress\""
    " option is a good choice for real-time video "
    "capture - it is faster than DivX ;-) and compresses "
    "data better than Motion JPEG. \n\n"
    "Home page:<a href=\"http://www.ligos.com/indeo\"> http://www.ligos.com/indeo</a>";

    char dvsd_about[] =
      "Sony Digital Video";

    char mjpg_about[] = 
    "Very fast Motion JPEG video codec, by Morgan"
    " Multimedia company. Recommended for"
    " video capture on slow machines or in high"
    " resolutions. Shareware. Current version"
    " is time-limited and will stop working after"
    " Nov. 1, 2000. Registration costs $25. Visit"
    " their web site for details. \n\n"
    " Web site: <a href=\"http://www.morgan-multimedia.com\">http://www.morgan-multimedia.com</a>";
    char mpg4_about[] =
    "Old beta version of Microsoft"
    " MPEG-4 video codec, incompatible"
    " with DivX ;-).";
    char cvid_about[] =
    "Very old video codec, usually"
    " available as a part of Windows.";
    char none_about[] =
    "No help available for this codec.";

    video_codecs.clear();

    vector<AttributeInfo> vs;
    vs.push_back("BitRate");
    vs.push_back("Crispness");
    vs.push_back("KeyFrames");
    vector<AttributeInfo> ds;
    ds.push_back(AttributeInfo("Quality", "Quality/CPU balance ( 0 fastest )", AttributeInfo::Integer, 0, 4));

    vector<AttributeInfo> ds1;
    ds1.push_back(AttributeInfo("Quality", "Quality/CPU balance ( 0 fastest )", AttributeInfo::Integer, 0, 4));
    ds1.push_back(AttributeInfo("Saturation", "Saturation", AttributeInfo::Integer, 0, 100));
    ds1.push_back(AttributeInfo("Hue", "Hue", AttributeInfo::Integer, 0, 100));
    ds1.push_back(AttributeInfo("Brightness", "Brightness", AttributeInfo::Integer, 0, 100));
    ds1.push_back(AttributeInfo("Contrast", "Contrast", AttributeInfo::Integer, 0, 100));
    
    video_codecs.push_back(CodecInfo(div3_codecs, "DivX ;-) low-motion DirectShow","divx_c32.ax", divx_about, CodecInfo::DShow_Dec, &CLSID_DivxDecompressorCF, vs, ds1));
    video_codecs.push_back(CodecInfo(wmv1_codecs, "Windows Media Video 7 DirectShow","wmvds32.ax", none_about, CodecInfo::DShow_Dec, &wmv1_clsid, vs, ds1));
    video_codecs.push_back(CodecInfo(wmv2_codecs, "Windows Media Video 7 DirectShow","wmv8ds32.ax", none_about, CodecInfo::DShow_Dec, &wmv2_clsid, vs, ds1));

    video_codecs.push_back(CodecInfo(dvsd_codecs, "DVSD", 	"qdv.dll",dvsd_about, CodecInfo::DShow_Dec, &dvsd_clsid, vs, ds1));


    video_codecs.push_back(CodecInfo(div3_codecs, "DivX ;-) low-motion","divxc32.dll", divx_about, CodecInfo::Win32, 0, vs, ds));
    video_codecs.push_back(CodecInfo(div4_codecs, "DivX ;-) fast-motion","divxc32f.dll",divx_about,CodecInfo::Win32, 0, vs, ds));
    
    vs.clear();
    ds.clear();
    ds.push_back(AttributeInfo("Saturation", "Saturation", AttributeInfo::Integer, -100, 100));
    ds.push_back(AttributeInfo("Brightness", "Brightness", AttributeInfo::Integer, -100, 100));
    ds.push_back(AttributeInfo("Contrast", "Contrast", AttributeInfo::Integer, -100, 100));
    video_codecs.push_back(CodecInfo(iv50_codecs, "Indeo Video 5.0 DirectShow", "ir50_32.dll", ivxx_about, CodecInfo::DShow_Dec, &CLSID_IV50_Decoder, vs, ds));

    vs.push_back("QuickCompress");
    video_codecs.push_back(CodecInfo(iv50_codecs, "Indeo Video 5.0",	"ir50_32.dll", ivxx_about, CodecInfo::Win32, 0, vs));
    video_codecs.push_back(CodecInfo(iv41_codecs, "Indeo Video 4.1",	"ir41_32.dll", ivxx_about, CodecInfo::Win32, 0, vs));
    video_codecs.push_back(CodecInfo(iv32_codecs, "Indeo Video 3.2",	"ir32_32.dll", ivxx_about, CodecInfo::Win32));
    video_codecs.push_back(CodecInfo(iv31_codecs, "Indeo Video 3.1", 	"ir32_32.dll", ivxx_about, CodecInfo::Win32));
//    video_codecs.push_back(CodecInfo(apxx_codecs, "AngelPotion MPEG-4", "APmpg4v1.dll", none_about, CodecInfo::Win32));

    vs.clear();
    char* mjpeg_modes[]=
    {
	"Fast integer",
	"Integer",
	"Float",
	0
    };	
    vs.push_back(AttributeInfo("Mode", "Calculation precision mode", (const char**)mjpeg_modes));
    vs.push_back(AttributeInfo("LicenseKey", "License key", AttributeInfo::String));
    vs.push_back(AttributeInfo("UserName", "User name", AttributeInfo::String));

    video_codecs.push_back(CodecInfo(mjpg_codecs, "Motion JPEG", 	"m3jpeg32.dll",mjpg_about, CodecInfo::Win32, 0, vs));


    video_codecs.push_back(CodecInfo(mpg4_codecs, "Microsoft MPEG-4", 	"mpg4c32.dll", mpg4_about, CodecInfo::Win32));
    video_codecs.push_back(CodecInfo(cvid_codecs, "Cinepak Video", 	"iccvid.dll" , cvid_about, CodecInfo::Win32));
    video_codecs.push_back(CodecInfo(vcr2_codecs, "ATI VCR-2", 		"ativcr2.dll", none_about, CodecInfo::Win32));
    video_codecs.push_back(CodecInfo(i263_codecs, "I263", 		"i263_32.drv", none_about, CodecInfo::Win32));
    video_codecs.push_back(CodecInfo(zero	, "Uncompressed", 	""	     , none_about, CodecInfo::Source));
    
//    video_codecs.push_back(CodecInfo(mjpg_codecs, "Motion JPEG",	"", CodecInfo::Source));
//    cerr<<"Registered all Win32 plugins"<<endl;
//
//	Here we can search for plug-in compressors/decompressors and insert them into array
//
    struct dirent **namelist;
    int n;
    n = scandir(def_path, &namelist, 0, alphasort);
    if(n>0)
    while(n--)
    {
	char* name=namelist[n]->d_name;
//	cerr<<"Checking "<<endl;
//	cerr<<name<<endl;
	if(strlen(name)<4)continue;
//	if(!n)continue;
	if(strcmp(name+strlen(name)-3, ".so"))continue;
	string lfn=string(def_path)+"/"+name;
//	cerr<<lfn<<endl;
	void* handle=dlopen(lfn.c_str(), RTLD_LAZY);
	if(!handle)continue;
	if(dlsym(handle, "GetPluginVersion")){dlclose(handle); continue;}
	const CodecInfo& (*RegisterPlugin)()=
	    (const CodecInfo& (*)())dlsym(handle, "RegisterPlugin");
	if(!RegisterPlugin)
	{
	    dlclose(handle);
	    continue;
	}
	CodecInfo ci=RegisterPlugin();
	cerr<<"Found "<<ci.text<<" plug-in"<<endl;
	ci.dll=lfn;
	ci.kind=CodecInfo::Plugin;
	video_codecs.push_back(ci);
	dlclose(handle);
    }	    
}

const CodecInfo* CodecInfo::match(int codec, const CodecInfo* start=0)
{
    vector<CodecInfo>::const_iterator it;
    if(start==0)
	it=video_codecs.begin();
    else
	it=start;
    vector<int>::const_iterator iv;
    for(; it!=video_codecs.end(); it++)
    {
	if(start && (it==start))continue;	
	for(iv=it->fourcc_array.begin(); iv!=it->fourcc_array.end(); iv++)
	    if(codec==(*iv))
		return (const CodecInfo*)it;
    }
    return 0;
}
