#include <loader.h>
#include <image.h>
#include <math.h>
#include <stdlib.h>
#include <except.h>
#define __MODULE__ "Image base class"
int CImage::imageCounter=0;
template<class T> inline T min(T a, T b)
{
    if(a<b)return a;
    else return b;
}    

void CImage::fill_members()
{
//    if(BitCount(_info->biCompression))
//	_info->biBitCount=BitCount(_info->biCompression);
    _yuv=_info->IsYUV();
    _depth=_info->biBitCount;
    _bpp=(_depth+7)/8;
    if(_depth==15)
	_bpl=2*_info->biWidth;
    else
	_bpl=(_depth*_info->biWidth)/8;
    _bytes=abs(_bpl*_info->biHeight);
    _pixels=abs(_info->biWidth*_info->biHeight);
    _width=abs(_info->biWidth);
    _height=abs(_info->biHeight);
}
CImage::CImage(const BitmapInfo* header, unsigned char* data=0, bool copy=true)
    :_info(new BitmapInfo(header)),_refcount(1)
{
    fill_members();
    if(!copy)
    {
        _data=data;
        _data_owner=false;
    }
    else
    {
        _data=new unsigned char[_bytes];
        _data_owner=true;
        if(data)
	    memcpy(_data, data, _bytes);
    }
    register_image();
}

CImage::CImage(unsigned char* data, int width, int height)
    :_info(new BitmapInfo(width, height, 24)), _refcount(1)
{
    if(!data)throw FATAL("Invalid argument");
    fill_members();
    _data=new unsigned char[_bytes];
    _data_owner=true;
    memcpy(_data, data, _bytes);
    register_image();
}

CImage::CImage(const CImage* im, int depth=-1)
    :_info(new BitmapInfo(im->get_fmt())), _refcount(1)
{
    if(depth!=-1)
	_info->setBits(depth);
    fill_members();
    _data=new unsigned char[_bytes];
    _data_owner=true;
    if(depth<0)
        memcpy(_data, im->data(), _bytes);
    else
	convert(im->data(), im->get_fmt());
    register_image();
}

CImage::CImage(const BitmapInfo* header, unsigned char* data, 
	const BitmapInfo* from_fmt)
    :_info(new BitmapInfo(header)),_refcount(1)
{
    if(!data)throw FATAL("Invalid argument");
    fill_members();
    _data=new unsigned char[_bytes];
    _data_owner=true;
    try
    {
	convert(data, from_fmt);
    }
    catch(...)
    {	
	delete _data;
	throw;
    }	
    register_image();
}
CImage::CImage(const CImage* im, const BitmapInfo* header)
	:_refcount(1), _info(new BitmapInfo(header))
{
    if(!im || !header)throw FATAL("Invalid argument");
    
    fill_members();
    _data=new unsigned char[_bytes];
    _data_owner=true;
    try
    {
	convert(im->data(), im->get_fmt());
    }
    catch(...)
    {	
	delete _data;
	throw;
    }	
    register_image();
}
CImage::~CImage()
{
    if(_data_owner)delete _data;
    unregister_image();
    delete _info;
}

unsigned char* CImage::at(int i)
{
    if(i<0)return 0;
    if(i>_height)return 0;
    return _data+i*_bpl;
}
const unsigned char* CImage::at(int i) const
{
    if(i<0)return 0;
    if(i>_height)return 0;
    return _data+i*_bpl;
}
unsigned char* CImage::offset(int i)
{
    return _data+i*_bpp;
}
const unsigned char* CImage::offset(int i) const
{
    return _data+i*_bpp;
}
unsigned char* CImage::at(int i, int j)
{
    if(j<0)return 0;
    if(j>_height)return 0;
    if(i<0)return 0;
    if(i>_width)return 0;
    return _data+i*_bpp+j*_bpl;
}

struct lookuptable
{
    long y[256];
    long rv[256];
    long gv[256];
    long gu[256];
    long bu[256];
    lookuptable()
    {
        for(int i=0; i<256; i++)
        {
    	    y[i]=298*(i-16);
	    rv[i]=408*(i-128);
    	    gv[i]=-208*(i-128);
	    gu[i]=-100*(i-128);
    	    bu[i]=517*(i-128);
        }
    }
};
lookuptable t;

col::col(yuv YUV)
{
    int R,G,B;

    R=t.y[YUV.Y]+t.rv[YUV.Cr];
    G=t.y[YUV.Y]+t.gv[YUV.Cr]+t.gu[YUV.Cb];
    B=t.y[YUV.Y]+t.bu[YUV.Cb];
    if(B<0)B=0;
    if(G<0)G=0;
    if(R<0)R=0;
    if(B>0xff00)B=0xff00;
    if(G>0xff00)G=0xff00;
    if(R>0xff00)R=0xff00;
    r=R>>8;
    g=G>>8;
    b=B>>8;
}
yuv::yuv(col Col)
{
    int _Y,_Cr,_Cb;
    _Y=66*Col.r+129*Col.g+25*Col.b+0x1000;
    _Cr=112*Col.r-94*Col.g-18*Col.b+0x8000;
    _Cb=-38*Col.r-74*Col.g+112*Col.b+0x8000;
    if(_Y<0x1000)_Y=0x1000;
    if(_Cr<0x1000)_Cr=0x1000;
    if(_Cb<0x1000)_Cb=0x1000;

    if(_Y>0xef00)_Y=0xef00;
    if(_Cr>0xef00)_Cr=0xef00;
    if(_Cb>0xef00)_Cb=0xef00;

    Y=_Y>>8;
    Cr=_Cr>>8;
    Cb=_Cb>>8;
}


int col::CREF() const
{
	return (int(b)<<16)+(int(g)<<8)+int(r);
}
void CImage::ToYUV(int destfmt=0)
{
    struct yuv* src;
    src=(struct yuv*)_data+_width*_height-1;
    if(_yuv)return;
    if(_depth!=24)throw FATAL("Cannot convert non-24 bit image to YUV");
    if(destfmt!=0)throw FATAL("Unsupported destination format");
    for(int i=_width*_height-1; i>=0; i--)
    {
    	*src=*(col*)src;
    	src--;
    }
    _yuv=fccYUV;
    _info->biCompression=fccYUV;
//    _yuv_fmt=0;
}

void CImage::ToRGB()
{
    if(!_yuv)return;
    if(_depth!=24)return;
    col* src=(col*)_data+_width*_height-1;
    for(int i=_width*_height-1; i>=0; i--)
    {
    	*src=*(struct yuv*)src;
    	src--;
    }
    _yuv=0;
    _info->setBits(24);
}
void CImage::to_16(const unsigned char* from)
{
    for(int i=0; i<_height; i++)
	for(int j=0; j<_width; j++)
	{
	    const unsigned char* src=from+3*(i*_width+j);
	    short* dest=(short*)_data+(i*_width+j);
	    short val=0;
	    val|=(src[0]>>3)<<11;
	    val|=(src[1]>>2)<<5;
	    val|=(src[2]>>3);
	    *dest=val;
        }	
}
void CImage::to_32(const unsigned char* from)
{
for(int i=0; i<_height; i++)
    for(int j=0; j<_width; j++)
    {
	memcpy(_data+4*j+i*_bpl, from+3*j+3*i*_width, 3);
	_data[4*j+i*_bpl+3]=0;
    }
}

void CImage::Blur(int range, int from)
{
    if(range<=0)return;
    if(from<0)return;
    if(range<=from)return;
    const int limit=_width*_height*3;
    unsigned  lim=1<<(range+1);
    unsigned  ach=3*(1<<(from+1));
    unsigned  char* p=(unsigned char*)_data;
    char g_pstore[200];
    __asm__ __volatile__
    ("fsave (%0)\n\t"
    "emms\n\t"
    :
    :"r"(&g_pstore));
    
    while(ach<3*lim)
	{
//	cerr<<"ach "<<ach<<" "<<lim<<endl;
		__asm__ __volatile__ (
		//p %0
		//ach %1
		//limit %2
			"pushl %%eax\n\t"
			"pushl %%ebx\n\t"
			"pushl %%ecx\n\t"
			"pushl %%edx\n\t"
			"pushl %%edi\n\t"

			"movl %0, %%ebx\n\t"
			
			"movl %1, %%ecx\n\t"

			"movl %0, %%edx\n\t"
			"addl %2, %%edx\n\t"
			"subl %1, %%edx\n\t"

			"movl %%edx, %%edi\n\t"

			"pxor %%mm2, %%mm2\n\t"

			"xorl %%eax, %%eax\n\t"
			"xorl %%edx, %%edx\n\t"
//			align 8
"m0:\n\t"
			"movd (%%ebx), %%mm0\n\t"
//			"movd %%ebx(%%ecx), %%mm1 \n\t"
			"movd (%%ecx,%%ebx), %%mm1 \n\t"
			"punpcklbw %%mm2, %%mm1\n\t"
			"punpcklbw %%mm2, %%mm0\n\t"
			"paddusw %%mm1, %%mm0 \n\t"
			"psrlw $1, %%mm0\n\t"
			"packuswb %%mm2, %%mm0\n\t"
			"movd  %%mm0, (%%ebx)\n\t"

			"addl $4, %%ebx\n\t"
			"cmpl %%edi, %%ebx\n\t"
			"jb m0\n\t"

			"popl %%edi\n\t"
			"popl %%edx\n\t"
			"popl %%ecx\n\t"
			"popl %%ebx\n\t"
			"popl %%eax\n\t"
		:
		:"m"(p), "m"(ach), "m"(limit)
		);			
	
		ach*=_width;

		__asm__ __volatile__ (
		//p %0
		//ach %1
		//limit %2
		"	pushl %%eax\n\t"
			"pushl %%ebx\n\t"
			"pushl %%ecx\n\t"
			"pushl %%edx\n\t"
			"pushl %%edi\n\t"

			"movl %0, %%ebx\n\t"
			
			"movl %0, %%ebx\n\t"
			
			"movl %1, %%ecx\n\t"

			"movl %0, %%edx\n\t"
			"addl %2, %%edx\n\t"
			"subl %1, %%edx\n\t"

			"movl %%edx, %%edi\n\t"

			"pxor %%mm2, %%mm2\n\t"

			"xorl %%eax, %%eax\n\t"
			"xorl %%edx, %%edx\n\t"
//			align 8
"m1:\n\t"
			"movd (%%ebx), %%mm0\n\t"
			"movd (%%ebx,%%ecx), %%mm1 \n\t"
			"punpcklbw %%mm2, %%mm1\n\t"
			"punpcklbw %%mm2, %%mm0\n\t"
			"paddusw %%mm1, %%mm0 \n\t"
			"psrlw $1, %%mm0\n\t"
			"packuswb %%mm2, %%mm0\n\t"
			"movd  %%mm0, (%%ebx)\n\t"

			"addl $4, %%ebx\n\t"
			"cmpl %%edi, %%ebx\n\t"
			"jb m1\n\t"

			"popl %%edi\n\t"
			"popl %%edx\n\t"
			"popl %%ecx\n\t"
			"popl %%ebx\n\t"
			"popl %%eax\n\t"
		:
		:"m"(p), "m"(ach), "m"(limit)
		);
		ach/=_width;
		ach*=2;
	}
	__asm__ __volatile__ ("frstor (%0)\n\t": :"r"(&g_pstore));
}
void CImage::convert(const unsigned char* data, const BitmapInfo* fmt)
{
    if(!data)throw FATAL("Invalid argument");
    if(!fmt)throw FATAL("Invalid argument");
#warning FIXME
//    cerr<<"S/W conversion requested"<<endl;
//    cerr<<"Src: "<<fmt->bpp()<<"/"<<fmt->biCompression<<endl;
//    cerr<<"Dst: "<<_info->bpp()<<"/"<<_info->biCompression<<endl;
    if(fmt->bpp()==_info->bpp())
    {
	memcpy(_data, data, _bytes);
	return;
    }
    if(fmt->IsYUV() && !_info->IsYUV())
    {
	unsigned char* tmp=new unsigned char[pixels()*3];
	memcpy(tmp, data, pixels()*3);
	CImage im(fmt, tmp, false);
	im.ToRGB();
	if(fmt->bpp()==24)
	    switch(_info->bpp())
	    {
    	    case 16:
		to_16(data);
		delete tmp;
	        return;
	    case 32:
		to_32(data);
		delete tmp;
		return;    
	    }
    }
    if(fmt->IsRGB() && _info->IsRGB())
    {
	if(fmt->bpp()==24)
	    switch(_info->bpp())
	    {
    	    case 16:
		to_16(data);
	        return;
	    case 32:
		to_32(data);
		return;    
	    }
    }
    cerr<<"Unsupported"<<endl;
}
BitmapInfo* CImage::get_fmt(){return _info;}
const BitmapInfo* CImage::get_fmt() const{return _info;}

bool CImage::supported(const BITMAPINFOHEADER& head)
{
    if(head.biCompression==0)
    {
	switch(head.biBitCount)
	{
	case 16:
	case 24:
	case 32:
	    return true;
	default:
	    return false;
	}
    }
    if(head.biCompression==3)
    {
	switch(head.biBitCount)
	{
	case 16:
	    return true;
	default:
	    return false;
	}
    }
    switch(head.biCompression)
    {
    case fccYUY2:
    case fccYV12:
    case fccIYUV:
    case fccUYVY:
    case fccYVYU:
	return true;
    }
    return false;
}
int CImage::UnknownColorSpace(int csp)
{
    switch(csp)
    {
	case 0:
	case 3:
	case fccYUY2:
	case fccYV12:
	case fccIYUV:
	case fccUYVY:
	case fccYVYU:
	    return 0;
    }
    return 1;
}
int CImage::BitCount(int csp)
{
    switch(csp)
    {
	case fccYUY2:
	case fccUYVY:
	case fccYVYU:
	    return 16;
	case fccYV12:
	case fccIYUV:
	    return 12;
    }
    return 0;
}
void CImage::byteswap()
{
    if(_info->biCompression!=0)return;
    if(_info->biBitCount!=24)return;
    unsigned char* t=_data;
    for(int i=0; i<_pixels; i++)
    {
	unsigned char tmp=t[0];
	t[0]=t[2];
	t[2]=tmp;
	t+=3;
    }
}

void * CImage::getaddr()
{
  return (void *)&_data;
}
